/**
 * @file   utils.c
 * @brief  Misc. functions to format strings, extract or push data, etc.
 *         Implementation.
 * @author Alejandro Álvarez Ayllón <aalvarez@cern.ch>
 */
#include <apr_strings.h>
#include <apr_xml.h>
#include <ctype.h>
#include <dmlite/c/dmlite.h>
#include <dmlite/c/io.h>
#include <dmlite/c/utils.h>
#include <inttypes.h>
#include <sys/stat.h>

#include "utils.h"


dav_error* dav_shared_new_error(request_rec *r, dmlite_context *ctx, int http_code,
                                const char *format, ...)
{
  const char *ctx_msg = "";
  const char *msg     = "";

  if (ctx != NULL) {
    ctx_msg = dmlite_error(ctx);

    /* Map err_no to HTTP status code, if needed */
    if (http_code == 0) {
      int err_code = dmlite_errno(ctx);
      switch(err_code)
      {
        case ENOSYS:
          http_code = HTTP_NOT_IMPLEMENTED;
          break;
        
        case DMLITE_NO_REPLICAS:
        case ENOENT:
          http_code = HTTP_NOT_FOUND;
          break;
        
        case EEXIST:
        case EISDIR:
          http_code = HTTP_CONFLICT;
          break;
        
        case ENOTDIR:
          http_code = HTTP_NOT_FOUND;
          break;

        case EACCES:
        case DMLITE_NO_SUCH_USER:
        case DMLITE_NO_SUCH_GROUP:
        case DMLITE_NO_USER_MAPPING:
          http_code = HTTP_FORBIDDEN;
          break;
        
        case ENOSPC:
          http_code = HTTP_INSUFFICIENT_STORAGE;
          break;
          
#ifdef ECOMM
        case ECOMM:
#endif
        case EBUSY:
          http_code = HTTP_SERVICE_UNAVAILABLE;
          break;
          
        default:
          http_code = HTTP_INTERNAL_SERVER_ERROR;
      }
    }
  }

  /* If extra information has been specified, generate the string */
  if (format) {
    va_list extra;

    va_start(extra, format);
    msg = apr_pvsprintf(r->pool, format, extra);
    va_end(extra);
  }

  msg = apr_xml_quote_string(r->pool,
                             apr_psprintf(r->pool, "%s (%s)", msg, ctx_msg),
                             1);

  /* Set error note */
  apr_table_setn(r->notes, "error-notes", msg);

#if AP_SERVER_MAJORVERSION_NUMBER == 2 &&\
    AP_SERVER_MINORVERSION_NUMBER < 4
  return  dav_new_error(r->pool, http_code, 0, msg);
#else
  return  dav_new_error(r->pool, http_code, 0,
                        (http_code < 400) ? APR_SUCCESS : APR_EGENERAL,
                        msg);
#endif
}



int dav_shared_format_datetime(char *buffer, size_t maxlen, time_t tstamp,
                               unsigned format)
{
  struct tm tms;
  int       l;

  gmtime_r(&tstamp, &tms);

  switch (format) {
    case DAV_DPM_RFC2068:
      l = snprintf(buffer, maxlen,
                   "%s, %.2d %s %d %.2d:%.2d:%.2d GMT",
                   apr_day_snames[tms.tm_wday],
                   tms.tm_mday, apr_month_snames[tms.tm_mon ],
                   tms.tm_year + 1900,
                   tms.tm_hour, tms.tm_min, tms.tm_sec);
      break;
    case DAV_DPM_RFC3339:
    default:
      l = snprintf(buffer, maxlen,
                   "%d-%.2d-%.2dT%.2d:%.2d:%.2dZ",
                   tms.tm_year + 1900, tms.tm_mon + 1,  tms.tm_mday,
                   tms.tm_hour, tms.tm_min, tms.tm_sec);
  }

  return l;
}



char *dav_shared_mode_str(char *str, mode_t mode)
{
  str[10] = '\0';

  if (S_ISDIR(mode))
    str[0] = 'd';
  else if (S_ISLNK(mode))
    str[0] = 'l';
  else
    str[0] = '-';
  
  str[1] = (mode & S_IRUSR) ? 'r':'-';
  str[2] = (mode & S_IWUSR) ? 'w':'-';
  if (mode & S_ISUID)
    str[3] = (mode & S_IXUSR) ? 's':'S';
  else
    str[3] = (mode & S_IXUSR) ? 'x':'-';

  str[4] = (mode & S_IRGRP) ? 'r':'-';
  str[5] = (mode & S_IWGRP) ? 'w':'-';
  if (mode & S_ISGID)
    str[6] = (mode & S_IXGRP) ? 's':'S';
  else
    str[6] = (mode & S_IXGRP) ? 'x':'-';

  str[7] = (mode & S_IROTH) ? 'r':'-';
  str[8] = (mode & S_IWOTH) ? 'w':'-';
  if (mode & S_ISVTX)
    str[9] = (mode & S_IXOTH) ? 't':'T';
  else
    str[9] = (mode & S_IXOTH) ? 'x':'-';

  return str;
}



char *dav_shared_size_str(char *str, size_t max, size_t size)
{
  if (size > 1073741823)
    snprintf(str, max, "%.1fG", (float)size / 1073741824.0);
  else if (size > 1048575)
    snprintf(str, max, "%.1fM", (float)size / 1048576.0);
  else if (size > 1023)
    snprintf(str, max, "%.1fK", (float)size / 1024.0);
  else
    snprintf(str, max, "%zu", size);
  return str;
}



void dav_shared_add_response(dav_walk_resource *wres, dav_error *error)
{
  dav_response *resp;

  resp = apr_pcalloc(wres->pool, sizeof(dav_response));
  resp->href = apr_pstrdup(wres->pool, wres->resource->uri);
  resp->status = error->status;
  resp->desc   = error->desc;

  resp->next = wres->response;
  wres->response = resp;
}



apr_table_t *dav_shared_parse_query(apr_pool_t *pool, const char *query, unsigned *size)
{
  apr_table_t *table;
  char        *copy, *last, *t, *p;

  table = apr_table_make(pool, 0);
  *size = 0;

  if (!query)
    return table;

  copy  = apr_pstrdup(pool, query);

  /* Split by & (fields) */
  t = apr_strtok(copy, "&", &last);
  while (t) {
    /* Split by = (values) */
    p = strchr(t, '=');
    if (p == NULL) {
      p = "";
    }
    else {
      *p = '\0';
      ++p;
    }
    /* Now, in t we have the field, and in p the value*/
    apr_table_set(table, t, p);
    ++(*size);
    /* Next */
    t = apr_strtok(NULL, "&", &last);
  }

  return table;
}



char *dav_shared_build_url(apr_pool_t *pool, dmlite_url *url,
        const char *default_scheme, unsigned default_port)
{
  char buffer[1024];

  /* Change the scheme if it doesn't fit us */
  if (strncmp(url->scheme, "http", 4) != 0) {
    strncpy(url->scheme, default_scheme, sizeof(url->scheme));
    url->scheme[sizeof(url->scheme) - 1] = '\0';
    url->port = default_port;
  }

  /* This serializes the URL as we want */
  dmlite_url_serialize(url, buffer, sizeof(buffer));

  return apr_pstrdup(pool, buffer);
}



char *dav_shared_build_aggregation_url(apr_pool_t* pool,
                                       int n_replicas,
                                       const char** replicas,
                                       int64_t* replicas_ids,
                                       const char* forbidden_str,
                                       const char* notfound_str,
                                       int64_t forbidden_id,
                                       int64_t notfound_id)
{
  char *redirect = NULL;
  if (n_replicas > 0) {
    const char *remaining = "";
    const char *forbidden = "";
    const char *notfound  = "";
    int   i;

    /* Build remaining list of replicas */
    for (i = 1; i < n_replicas; ++i) {
      remaining = apr_psprintf(pool, "%s&r%d=%"PRId64",%s",
                               remaining, i, replicas_ids[i],
                               replicas[i]);
    }

    /* Append forbidden? */
    if (forbidden_id >= 0) {
      forbidden = apr_psprintf(pool, "&forbidden=%s%s%"PRId64,
                               forbidden_str?forbidden_str:"",
                               (forbidden_str && forbidden_str[0])?",":"",
                               forbidden_id);
    }
    else if (forbidden_str != NULL) {
      forbidden = apr_psprintf(pool, "&forbidden=%s",
                               forbidden_str);
    }

    /* Append not found? */
    if (notfound_id >= 0) {
      notfound = apr_psprintf(pool, "&notfound=%s%s%"PRId64,
                              notfound_str?notfound_str:"",
                              (notfound_str && notfound_str[0])?",":"",
                               notfound_id);
    }
    else if (notfound_str != NULL) {
      notfound = apr_psprintf(pool, "&notfound=%s",
                              notfound_str);
    }

    /* Redirect URL */
    redirect = apr_psprintf(pool, "%s?rid=%"PRId64"%s%s%s",
                            replicas[0],
                            replicas_ids[0],
                            forbidden,
                            notfound,
                            remaining);
  }
  return redirect;
}



int dav_shared_request_accepts(request_rec* r, const char* content)
{
  const char* p = apr_table_get(r->headers_in, "Accept");
  
  while (p) {
    while (isspace(*p) || ispunct(*p)) ++p;
    if (strncmp(content, p, 25) == 0)
      return 1;
    p = strchr(p, ',');
  }
  
  return 0;
}



apr_status_t dav_shared_context_free(void* context)
{
  dmlite_context_free((dmlite_context*)context);
  return APR_SUCCESS;
}



apr_status_t dav_shared_dict_free(void* dict)
{
  dmlite_any_dict_free((dmlite_any_dict*)dict);
  return APR_SUCCESS;
}



apr_status_t dav_shared_fclose(void* fd)
{
  dmlite_fclose((dmlite_fd*)fd);
  return APR_SUCCESS;
}
